"""Code that handles CLI commands to upload"""

import datetime as dt
import logging

import click

from cg.apps.tb import TrailblazerAPI
from cg.apps.tb.models import TrailblazerAnalysis
from cg.constants import EXIT_FAIL, EXIT_SUCCESS, Priority, Workflow
from cg.constants.constants import DRY_RUN
from cg.constants.tb import AnalysisTypes
from cg.meta.deliver.utils import get_delivery_scope
from cg.meta.rsync import RsyncAPI
from cg.store.models import Case
from cg.store.store import Store

LOG = logging.getLogger(__name__)


@click.command("clinical-delivery")
@click.pass_context
@click.argument("case_id", required=True)
@DRY_RUN
def upload_clinical_delivery(context: click.Context, case_id: str, dry_run: bool):
    """Links the appropriate files for a case, based on the data_delivery, to the customer folder
    and subsequently uses rsync to upload it to caesar."""

    click.echo(click.style("----------------- Clinical-delivery -----------------"))

    case: Case = context.obj.status_db.get_case_by_internal_id(case_id)
    workflows: set[str] = case.get_delivery_arguments()
    is_sample_delivery: bool
    is_case_delivery: bool
    is_complete_delivery: bool
    job_id: int
    is_sample_delivery, is_case_delivery = get_delivery_scope(workflows)
    if not workflows:
        LOG.info(f"No delivery of files requested for case {case_id}")
        return

    LOG.debug(f"Delivery types are: {workflows}")
    for delivery_type in workflows:
        context.obj.delivery_api.deliver_files(case=case, workflow=delivery_type)

    rsync_api = RsyncAPI(context.obj)
    is_complete_delivery, job_id = rsync_api.slurm_rsync_single_case(
        case=case,
        dry_run=dry_run,
        sample_files_present=is_sample_delivery,
        case_files_present=is_case_delivery,
    )
    RsyncAPI.write_trailblazer_config(
        {"jobs": [str(job_id)]}, config_path=rsync_api.trailblazer_config_path
    )
    analysis_name: str = f"{case_id}_rsync" if is_complete_delivery else f"{case_id}_partial"
    if not dry_run:
        trailblazer_api: TrailblazerAPI = context.obj.trailblazer_api
        analysis: TrailblazerAnalysis = trailblazer_api.add_pending_analysis(
            case_id=analysis_name,
            analysis_type=AnalysisTypes.OTHER,
            config_path=rsync_api.trailblazer_config_path.as_posix(),
            out_dir=rsync_api.log_dir.as_posix(),
            slurm_quality_of_service=Priority.priority_to_slurm_qos().get(case.priority),
            workflow=Workflow.RSYNC,
            ticket=case.latest_ticket,
        )
        trailblazer_api.add_upload_job_to_analysis(analysis_id=analysis.id, slurm_id=job_id)
    LOG.info(f"Transfer of case {case_id} started with SLURM job id {job_id}")


@click.command("all-fastq")
@click.pass_context
@DRY_RUN
def auto_fastq(context: click.Context, dry_run: bool):
    """Starts upload of all not previously uploaded cases with analysis type fastq to
    clinical-delivery."""
    exit_code: int = EXIT_SUCCESS
    status_db: Store = context.obj.status_db
    trailblazer_api: TrailblazerAPI = context.obj.trailblazer_api
    for analysis_obj in status_db.get_analyses_to_upload(workflow=Workflow.FASTQ):
        if analysis_obj.case.analyses[0].uploaded_at:
            LOG.debug(
                f"Newer analysis already uploaded for {analysis_obj.case.internal_id}, skipping"
            )
            continue
        if analysis_obj.upload_started_at:
            if trailblazer_api.is_latest_analysis_completed(case_id=analysis_obj.case.internal_id):
                LOG.info(
                    f"The upload for {analysis_obj.case.internal_id} is completed, setting uploaded at to {dt.datetime.now()}"
                )
                analysis_obj.uploaded_at = dt.datetime.now()
            else:
                LOG.debug(
                    f"Upload to clinical-delivery for {analysis_obj.case.internal_id} has already started, skipping"
                )
            continue
        case: Case = analysis_obj.case
        LOG.info(f"Uploading family: {case.internal_id}")
        analysis_obj.upload_started_at = dt.datetime.now()
        try:
            context.invoke(upload_clinical_delivery, case_id=case.internal_id, dry_run=dry_run)
        except Exception as error:
            LOG.error(f"Upload of case {case.internal_id} failed")
            LOG.error(error)
            exit_code: int = EXIT_FAIL
            continue

    if not dry_run:
        status_db.session.commit()

    if exit_code == EXIT_FAIL:
        raise click.Abort



import logging
import os
from pathlib import Path
from cg.constants import delivery as constants
from cg.constants.constants import Workflow
from cg.store.models import Case, Sample
from housekeeper.store.models import File

LOG = logging.getLogger(__name__)


def get_delivery_scope(workflows: list[str]) -> tuple[bool, bool]:
    """Returns the scope of the delivery, ie whether sample and/or case files were delivered."""
    case_delivery: bool = False
    sample_delivery: bool = False
    for workflow in workflows:
        if is_sample_delivery(workflow):
            sample_delivery = True
        if is_case_delivery(workflow):
            case_delivery = True
    return sample_delivery, case_delivery


def is_sample_delivery(workflow: str) -> bool:
    return get_sample_tags_for_workflow(workflow) and workflow_has_one_case_per_ticket(workflow)


def is_case_delivery(workflow: str) -> bool:
    return get_case_tags_for_workflow(workflow)


def get_delivery_dir_path(
    base_path: Path,
    customer_id: str,
    ticket: str,
    case_name: str = None,
    sample_name: str = None,
) -> Path:
    """Get a path for delivering files.

    Note that case name and sample name needs to be the identifiers sent from customer.
    """
    delivery_path = Path(base_path, customer_id, constants.INBOX_NAME, ticket)
    if case_name:
        delivery_path = Path(delivery_path, case_name)
    if sample_name:
        delivery_path = Path(delivery_path, sample_name)
    return delivery_path


def workflow_has_one_case_per_ticket(workflow: Workflow) -> bool:
    return workflow in constants.ONLY_ONE_CASE_PER_TICKET


def get_case_tags_for_workflow(workflow: Workflow) -> list[set[str]]:
    return constants.PIPELINE_ANALYSIS_TAG_MAP[workflow]["case_tags"]


def get_sample_tags_for_workflow(workflow: Workflow) -> list[set[str]]:
    return constants.PIPELINE_ANALYSIS_TAG_MAP[workflow]["sample_tags"]


def get_delivery_case_name(case: Case, workflow: str) -> str | None:
    return None if workflow in constants.ONLY_ONE_CASE_PER_TICKET else case.name


def get_out_path(out_dir: Path, file: Path, case: Case) -> Path:
    out_file_name: str = _get_case_out_file_name(file=file, case=case)
    return Path(out_dir, out_file_name)


def _get_case_out_file_name(file: Path, case: Case) -> str:
    return file.name.replace(case.internal_id, case.name)


def get_sample_out_file_name(file: Path, sample: Sample) -> str:
    return file.name.replace(sample.internal_id, sample.name)


def should_include_file_case(file: File, sample_ids: set[str], workflow: str) -> bool:
    """Check if file should be included in case bundle.

    At least one tag should match between file and tags.
    Do not include files with sample tags.
    """
    tags_on_file = {tag.name for tag in file.tags}
    case_tags = get_case_tags_for_workflow(workflow)

    if sample_ids.intersection(tags_on_file):
        return False

    return any(tags.issubset(tags_on_file) for tags in case_tags)


def should_include_file_sample(file: File, workflow: str) -> bool:
    """Check if file should be included in sample bundle.

    At least one tag should match between file and the sample tags.
    """
    file_tags = {tag.name for tag in file.tags}
    sample_tags = get_sample_tags_for_workflow(workflow)
    return any(tags.issubset(file_tags) for tags in sample_tags)


def create_link(source: Path, destination: Path, dry_run: bool = False) -> bool:
    if dry_run:
        LOG.info(f"Would hard link file {source} to {destination}")
        return True
    try:
        os.link(source, destination)
        LOG.info(f"Hard link file {source} to {destination}")
        return True
    except FileExistsError:
        LOG.info(f"Path {destination} exists, skipping")
        return False


def get_bundle_name(case: Case, sample: Sample, workflow: str) -> str:
    return sample.internal_id if workflow == Workflow.FASTQ else case.internal_id



from pathlib import Path
import pytest

from cg.services.fastq_file_service.fastq_file_service import FastqFileService


@pytest.fixture
def fastq_file_service():
    return FastqFileService()


def create_fastqs_directory(number_forward_reads, number_reverse_reads, tmp_path):
    fastq_dir = Path(tmp_path, "fastqs")
    fastq_dir.mkdir()
    for i in range(number_forward_reads):
        file = Path(fastq_dir, f"sample_R1_{i}.fastq.gz")
        file.write_text(f"forward read {i}")

    for i in range(number_reverse_reads):
        file = Path(fastq_dir, f"sample_R2_{i}.fastq.gz")
        file.write_text(f"reverse read {i}")
    return fastq_dir


@pytest.fixture
def fastqs_dir(tmp_path) -> Path:
    return create_fastqs_directory(
        number_forward_reads=3, number_reverse_reads=3, tmp_path=tmp_path
    )


@pytest.fixture
def fastqs_forward(tmp_path) -> Path:
    """Return a directory with only forward reads."""
    return create_fastqs_directory(
        number_forward_reads=3, number_reverse_reads=0, tmp_path=tmp_path
    )



"""Module to parse the BCL convert metrics data into the sequencing statistics model."""

from datetime import datetime
from pathlib import Path

from cg.apps.sequencing_metrics_parser.parsers.bcl_convert import (
    BclConvertMetricsParser,
)
from cg.apps.sequencing_metrics_parser.parsers.utils import get_flow_cell_id
from cg.constants.demultiplexing import UNDETERMINED
from cg.store.models import SampleLaneSequencingMetrics


def create_sample_lane_sequencing_metrics_from_bcl_convert_metrics_for_flow_cell(
    flow_cell_dir: Path,
) -> list[SampleLaneSequencingMetrics]:
    """Parse the BCL convert metrics data into the sequencing statistics model."""
    metrics_parser: BclConvertMetricsParser = BclConvertMetricsParser(
        bcl_convert_metrics_dir_path=flow_cell_dir,
    )
    sample_internal_ids: list[str] = metrics_parser.get_sample_internal_ids()
    sample_lane_sequencing_metrics: list[SampleLaneSequencingMetrics] = []

    for sample_internal_id in sample_internal_ids:
        for lane in metrics_parser.get_lanes_for_sample(sample_internal_id=sample_internal_id):
            metrics: SampleLaneSequencingMetrics = create_bcl_convert_sequencing_metrics(
                sample_internal_id=sample_internal_id, lane=lane, metrics_parser=metrics_parser
            )
            sample_lane_sequencing_metrics.append(metrics)
    return sample_lane_sequencing_metrics


def create_bcl_convert_undetermined_metrics(
    flow_cell_dir: Path, non_pooled_lane_sample_pairs: list[tuple[int, str]]
) -> list[SampleLaneSequencingMetrics]:
    """Return sequencing metrics for any undetermined reads in the specified lanes."""
    metrics_parser = BclConvertMetricsParser(flow_cell_dir)
    undetermined_metrics: list[SampleLaneSequencingMetrics] = []

    for lane, sample_internal_id in non_pooled_lane_sample_pairs:
        if not metrics_parser.has_undetermined_reads_in_lane(lane):
            continue

        # Passing Undetermined as the sample id is required to extract the undetermined reads data.
        # BclConvert tags undetermined reads in a lane with the sample id "Undetermined".
        metrics: SampleLaneSequencingMetrics = create_bcl_convert_sequencing_metrics(
            sample_internal_id=UNDETERMINED, lane=lane, metrics_parser=metrics_parser
        )
        metrics.sample_internal_id = sample_internal_id
        undetermined_metrics.append(metrics)
    return undetermined_metrics


def create_bcl_convert_sequencing_metrics(
    sample_internal_id: str, lane: int, metrics_parser: BclConvertMetricsParser
) -> SampleLaneSequencingMetrics:
    """Create sequencing metrics for a sample in a lane."""
    flow_cell_id: str = get_flow_cell_id(metrics_parser.bcl_convert_demultiplex_dir.name)

    total_reads: int = metrics_parser.calculate_total_reads_for_sample_in_lane(
        sample_internal_id=sample_internal_id, lane=lane
    )
    q30_bases_percent: float = metrics_parser.get_q30_bases_percent_for_sample_in_lane(
        sample_internal_id=sample_internal_id, lane=lane
    )
    mean_quality_score: float = metrics_parser.get_mean_quality_score_for_sample_in_lane(
        sample_internal_id=sample_internal_id, lane=lane
    )
    return SampleLaneSequencingMetrics(
        sample_internal_id=sample_internal_id,
        flow_cell_name=flow_cell_id,
        flow_cell_lane_number=lane,
        sample_total_reads_in_lane=total_reads,
        sample_base_percentage_passing_q30=q30_bases_percent,
        sample_base_mean_quality_score=mean_quality_score,
        created_at=datetime.now(),
    )



"""Test the cli for uploading using auto"""

import datetime
import logging

from click.testing import CliRunner

from cg.cli.upload.base import upload_all_completed_analyses
from cg.constants import Workflow
from cg.models.cg_config import CGConfig
from tests.store_helpers import StoreHelpers


def test_upload_auto_with_workflow(
    cli_runner: CliRunner,
    helpers: StoreHelpers,
    timestamp: datetime.datetime,
    upload_context: CGConfig,
    caplog,
):
    """Test upload auto"""
    # GIVEN a store with a workflow
    workflow = Workflow.MIP_DNA
    helpers.add_analysis(store=upload_context.status_db, completed_at=timestamp, workflow=workflow)

    # WHEN uploading all analysis from workflow MIP
    caplog.set_level(logging.INFO)
    cli_runner.invoke(upload_all_completed_analyses, ["--workflow", "mip-dna"], obj=upload_context)
    # THEN assert that the MIP analysis was successfully uploaded
    assert "Uploading analysis for case" in caplog.text



from datetime import datetime
from enum import Enum
from typing import Callable

from sqlalchemy import and_, not_, or_
from sqlalchemy.orm import Query

from cg.constants import REPORT_SUPPORTED_DATA_DELIVERY
from cg.constants.constants import CaseActions, DataDelivery, Workflow
from cg.constants.observations import (
    LOQUSDB_BALSAMIC_SEQUENCING_METHODS,
    LOQUSDB_MIP_SEQUENCING_METHODS,
    LOQUSDB_SUPPORTED_WORKFLOWS,
)
from cg.store.models import Analysis, Application, Case, Customer, Sample


def filter_cases_by_action(cases: Query, action: str, **kwargs) -> Query:
    """Filter cases with matching action."""
    return cases.filter(Case.action == action) if action else cases


def filter_cases_by_case_search(cases: Query, case_search: str, **kwargs) -> Query:
    """Filter cases with matching internal id or name."""
    return (
        cases.filter(
            or_(
                Case.internal_id.contains(case_search),
                Case.name.contains(case_search),
            )
        )
        if case_search
        else cases
    )


def filter_cases_by_customer_entry_id(cases: Query, customer_entry_id: int, **kwargs) -> Query:
    """Filter cases with matching customer id."""
    return cases.filter(Case.customer_id == customer_entry_id)


def filter_cases_by_customer_entry_ids(
    cases: Query, customer_entry_ids: list[int], **kwargs
) -> Query:
    """Filter cases with matching customer ids."""
    return cases.filter(Case.customer_id.in_(customer_entry_ids)) if customer_entry_ids else cases


def filter_cases_by_entry_id(cases: Query, entry_id: int, **kwargs) -> Query:
    """Filter cases by entry id."""
    return cases.filter(Case.id == entry_id)


def filter_case_by_internal_id(cases: Query, internal_id: str, **kwargs) -> Query:
    """Filter cases with matching internal id."""
    return cases.filter(Case.internal_id == internal_id)


def filter_cases_by_internal_id_search(cases: Query, internal_id_search: str, **kwargs) -> Query:
    """Filter cases with internal ids matching the search pattern."""
    return cases.filter(Case.internal_id.contains(internal_id_search))


def filter_cases_by_name(cases: Query, name: str, **kwargs) -> Query:
    """Filter cases with matching name."""
    return cases.filter(Case.name == name) if name else cases


def filter_cases_by_name_search(cases: Query, name_search: str, **kwargs) -> Query:
    """Filter cases with names matching the search pattern."""
    return cases.filter(Case.name.contains(name_search))


def filter_cases_by_workflow_search(cases: Query, workflow_search: str, **kwargs) -> Query:
    """Filter cases with a workflow search pattern."""
    return cases.filter(Case.data_analysis.contains(workflow_search))


def filter_cases_by_priority(cases: Query, priority: str, **kwargs) -> Query:
    """Filter cases with matching priority."""
    return cases.filter(Case.priority == priority)


def filter_cases_by_ticket_id(cases: Query, ticket_id: str, **kwargs) -> Query:
    """Filter cases with matching ticket id."""
    return cases.filter(Case.tickets.contains(ticket_id))


def filter_cases_for_analysis(cases: Query, **kwargs) -> Query:
    """Filter cases in need of analysis by:
    1. Action set to analyze or
    2. Internally created cases with no action set and no prior analysis or
    3. Cases with no action, but new sequence data
    """
    return cases.filter(
        or_(
            Case.action == CaseActions.ANALYZE,
            and_(
                Application.is_external.isnot(True),
                Case.action.is_(None),
                Analysis.created_at.is_(None),
            ),
            and_(
                Case.action.is_(None),
                Analysis.created_at < Sample.last_sequenced_at,
            ),
        )
    )


def filter_cases_has_sequence(cases: Query, **kwargs) -> Query:
    """Filter cases that are not sequenced according to record in StatusDB."""
    return cases.filter(or_(Application.is_external, Sample.last_sequenced_at.isnot(None)))


def filter_cases_not_analysed(cases: Query, **kwargs) -> Query:
    """Filter cases that have not been analysed and are not currently being analysed."""
    not_analyzed_condition = not_(Case.analyses.any(Analysis.completed_at.isnot(None)))
    not_in_progress_condition = Case.action != CaseActions.ANALYZE

    return cases.filter(and_(not_analyzed_condition, not_in_progress_condition))


def filter_cases_with_workflow(cases: Query, workflow: Workflow = None, **kwargs) -> Query:
    """Filter cases with workflow."""
    return cases.filter(Case.data_analysis == workflow) if workflow else cases


def filter_cases_with_loqusdb_supported_workflow(
    cases: Query, workflow: Workflow = None, **kwargs
) -> Query:
    """Filter Loqusdb related cases with workflow."""
    records: Query = (
        cases.filter(Case.data_analysis == workflow)
        if workflow
        else cases.filter(Case.data_analysis.in_(LOQUSDB_SUPPORTED_WORKFLOWS))
    )
    return records.filter(Customer.loqus_upload == True)


def filter_cases_with_loqusdb_supported_sequencing_method(
    cases: Query, workflow: Workflow = None, **kwargs
) -> Query:
    """Filter cases with Loqusdb supported sequencing method."""
    supported_sequencing_methods = {
        Workflow.MIP_DNA: LOQUSDB_MIP_SEQUENCING_METHODS,
        Workflow.BALSAMIC: LOQUSDB_BALSAMIC_SEQUENCING_METHODS,
    }
    return (
        cases.filter(Application.prep_category.in_(supported_sequencing_methods[workflow]))
        if workflow
        else cases
    )


def filter_cases_with_scout_data_delivery(cases: Query, **kwargs) -> Query:
    """Filter cases containing Scout as a data delivery option."""
    return cases.filter(Case.data_delivery.contains(DataDelivery.SCOUT))


def filter_newer_cases_by_order_date(cases: Query, order_date: datetime, **kwargs) -> Query:
    """Filter cases newer than date."""
    cases: Query = cases.filter(Case.ordered_at > order_date)
    return cases.order_by(Case.ordered_at.asc())


def filter_inactive_analysis_cases(cases: Query, **kwargs) -> Query:
    """Filter cases which are not set or on hold."""
    return cases.filter(
        or_(
            Case.action.is_(None),
            Case.action == CaseActions.HOLD,
        )
    )


def filter_older_cases_by_creation_date(cases: Query, creation_date: datetime, **kwargs) -> Query:
    """Filter older cases compared to date."""
    cases = cases.filter(Case.created_at < creation_date)
    return cases.order_by(Case.created_at.asc())


def filter_report_supported_data_delivery_cases(cases: Query, **kwargs) -> Query:
    """Filter cases with a valid data delivery for delivery report generation."""
    return cases.filter(Case.data_delivery.in_(REPORT_SUPPORTED_DATA_DELIVERY))


def filter_running_cases(cases: Query, **kwargs) -> Query:
    """Filter cases which are running."""
    return cases.filter(Case.action == CaseActions.RUNNING)


def filter_compressible_cases(cases: Query, **kwargs) -> Query:
    """Filter cases which are running."""
    return cases.filter(Case.is_compressible)


def order_cases_by_created_at(cases: Query, **kwargs) -> Query:
    """Order cases by created at."""
    return cases.order_by(Case.created_at.desc())


def apply_case_filter(
    cases: Query,
    filter_functions: list[Callable],
    action: str | None = None,
    case_search: str | None = None,
    creation_date: datetime | None = None,
    customer_entry_id: int | None = None,
    customer_entry_ids: list[int] | None = None,
    entry_id: int | None = None,
    internal_id: str | None = None,
    internal_id_search: str | None = None,
    name: str | None = None,
    name_search: str | None = None,
    order_date: datetime | None = None,
    workflow: Workflow | None = None,
    workflow_search: str | None = None,
    priority: str | None = None,
    ticket_id: str | None = None,
) -> Query:
    """Apply filtering functions and return filtered results."""
    for function in filter_functions:
        cases: Query = function(
            action=action,
            case_search=case_search,
            cases=cases,
            creation_date=creation_date,
            customer_entry_id=customer_entry_id,
            customer_entry_ids=customer_entry_ids,
            entry_id=entry_id,
            internal_id=internal_id,
            internal_id_search=internal_id_search,
            name=name,
            name_search=name_search,
            order_date=order_date,
            workflow=workflow,
            workflow_search=workflow_search,
            priority=priority,
            ticket_id=ticket_id,
        )
    return cases


class CaseFilter(Enum):
    """Define case filters."""

    FILTER_BY_ACTION: Callable = filter_cases_by_action
    FILTER_BY_CASE_SEARCH: Callable = filter_cases_by_case_search
    FILTER_BY_CUSTOMER_ENTRY_ID: Callable = filter_cases_by_customer_entry_id
    FILTER_BY_CUSTOMER_ENTRY_IDS: Callable = filter_cases_by_customer_entry_ids
    FILTER_BY_ENTRY_ID: Callable = filter_cases_by_entry_id
    FILTER_BY_INTERNAL_ID: Callable = filter_case_by_internal_id
    FILTER_BY_INTERNAL_ID_SEARCH: Callable = filter_cases_by_internal_id_search
    FILTER_BY_NAME: Callable = filter_cases_by_name
    FILTER_BY_NAME_SEARCH: Callable = filter_cases_by_name_search
    FILTER_BY_WORKFLOW_SEARCH: Callable = filter_cases_by_workflow_search
    FILTER_BY_PRIORITY: Callable = filter_cases_by_priority
    FILTER_BY_TICKET: Callable = filter_cases_by_ticket_id
    FILTER_FOR_ANALYSIS: Callable = filter_cases_for_analysis
    FILTER_HAS_INACTIVE_ANALYSIS: Callable = filter_inactive_analysis_cases
    FILTER_HAS_SEQUENCE: Callable = filter_cases_has_sequence
    FILTER_IS_RUNNING: Callable = filter_running_cases
    FILTER_IS_COMPRESSIBLE: Callable = filter_compressible_cases
    FILTER_NEW_BY_ORDER_DATE: Callable = filter_newer_cases_by_order_date
    FILTER_NOT_ANALYSED: Callable = filter_cases_not_analysed
    FILTER_OLD_BY_CREATION_DATE: Callable = filter_older_cases_by_creation_date
    FILTER_REPORT_SUPPORTED: Callable = filter_report_supported_data_delivery_cases
    FILTER_WITH_LOQUSDB_SUPPORTED_WORKFLOW: Callable = filter_cases_with_loqusdb_supported_workflow
    FILTER_WITH_LOQUSDB_SUPPORTED_SEQUENCING_METHOD: Callable = (
        filter_cases_with_loqusdb_supported_sequencing_method
    )
    FILTER_WITH_WORKFLOW: Callable = filter_cases_with_workflow
    FILTER_WITH_SCOUT_DELIVERY: Callable = filter_cases_with_scout_data_delivery
    ORDER_BY_CREATED_AT: Callable = order_cases_by_created_at



import logging
from pathlib import Path

from housekeeper.store.models import Version
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import delivery as constants
from cg.constants.constants import DataDelivery, Workflow
from cg.exc import MissingFilesError
from cg.meta.deliver.utils import (
    create_link,
    get_bundle_name,
    get_delivery_case_name,
    get_delivery_dir_path,
    get_case_tags_for_workflow,
    get_out_path,
    get_sample_out_file_name,
    get_sample_tags_for_workflow,
    should_include_file_case,
    should_include_file_sample,
)
from cg.services.fastq_file_service.fastq_file_service import FastqFileService
from cg.store.store import Store
from cg.store.models import Case, Sample

LOG = logging.getLogger(__name__)


class DeliveryAPI:
    """In this module, delivery means linking files in a bundle to the customers folder on the compute cluster."""

    def __init__(
        self,
        store: Store,
        hk_api: HousekeeperAPI,
        customers_folder: Path,
        fastq_file_service: FastqFileService,
        force_all: bool = False,
        ignore_missing_bundles: bool = False,
        dry_run: bool = False,
    ):
        self.store = store
        self.hk_api = hk_api
        self.customers_folder = customers_folder
        self.ignore_missing_bundles = ignore_missing_bundles
        self.deliver_failed_samples = force_all
        self.dry_run = dry_run
        self.fastq_file_service = fastq_file_service

    def deliver(self, ticket: str | None, case_id: str | None, workflow: str) -> None:
        if ticket:
            self._deliver_files_by_ticket(ticket=ticket, workflow=workflow)
        elif case_id:
            self._deliver_files_by_case(case_id=case_id, workflow=workflow)

    def _deliver_files_by_ticket(self, ticket: str, workflow: str) -> None:
        cases: list[Case] = self.store.get_cases_by_ticket_id(ticket)
        if not cases:
            LOG.warning(f"Could not find cases for ticket {ticket}")
            return
        for case in cases:
            self.deliver_files(case=case, workflow=workflow)

    def _deliver_files_by_case(self, case_id: str, workflow: str) -> None:
        case: Case | None = self.store.get_case_by_internal_id(case_id)
        if not case:
            LOG.warning(f"Could not find case {case_id}")
            return
        self.deliver_files(case=case, workflow=workflow)

    def deliver_files(self, case: Case, workflow: str):
        if self._is_case_deliverable(case=case, workflow=workflow):
            self._deliver_case_files(case=case, workflow=workflow)
            self._deliver_sample_files(case=case, workflow=workflow)

    def _deliver_case_files(self, case: Case, workflow: str) -> None:
        LOG.debug(f"Deliver case files for {case.internal_id}")
        if not get_case_tags_for_workflow(workflow):
            return
        self._link_case_files(case=case, workflow=workflow)

    def _link_case_files(self, case: Case, workflow: str):
        out_dir: Path = self._create_case_delivery_directory(case)
        files: list[Path] = self._get_case_files(workflow=workflow, case=case)
        for file in files:
            out_file: Path = get_out_path(out_dir=out_dir, file=file, case=case)
            create_link(source=file, destination=out_file, dry_run=self.dry_run)

    def _create_case_delivery_directory(self, case: Case) -> Path:
        case_target_directory: Path = get_delivery_dir_path(
            case_name=case.name,
            customer_id=case.customer.internal_id,
            ticket=case.latest_ticket,
            base_path=self.customers_folder,
        )
        if not self.dry_run:
            LOG.debug(f"Creating delivery directory for case {case_target_directory}")
            case_target_directory.mkdir(parents=True, exist_ok=True)
        return case_target_directory

    def _deliver_sample_files(self, case: Case, workflow: str) -> None:
        if not get_sample_tags_for_workflow(workflow):
            return

        deliverable_samples: list[Sample] = self._get_deliverable_samples(
            case=case, workflow=workflow
        )
        case_name: str | None = get_delivery_case_name(case=case, workflow=workflow)
        for sample in deliverable_samples:
            sample_directory: Path = self._create_sample_delivery_directory(
                case=case, sample=sample, workflow=workflow
            )
            number_linked_files: int = 0
            version: Version = self._get_version_for_sample(
                sample=sample, case=case, workflow=workflow
            )
            files: list[Path] = self._get_sample_files_from_version(
                version=version, workflow=workflow
            )
            for file_path in files:
                file_name: str = get_sample_out_file_name(file=file_path, sample=sample)
                if case_name:
                    file_name: str = file_name.replace(case.internal_id, case.name)
                out_path = Path(sample_directory, file_name)
                if create_link(source=file_path, destination=out_path, dry_run=self.dry_run):
                    number_linked_files += 1

            if not files and number_linked_files == 0:
                raise MissingFilesError(f"Could not find any files for sample {sample.internal_id}")
            if workflow == Workflow.FASTQ and case.data_analysis == Workflow.MICROSALT:
                self.concatenate_fastqs(sample_directory=sample_directory, sample_name=sample.name)

    def _create_sample_delivery_directory(self, case: Case, sample: Sample, workflow: str) -> Path:
        case_name: str | None = get_delivery_case_name(case=case, workflow=workflow)
        sample_directory: Path = get_delivery_dir_path(
            case_name=case_name,
            sample_name=sample.name,
            customer_id=case.customer.internal_id,
            ticket=case.latest_ticket,
            base_path=self.customers_folder,
        )
        if not self.dry_run:
            LOG.debug(f"Creating sample delivery directory {sample_directory}")
            sample_directory.mkdir(parents=True, exist_ok=True)
        return sample_directory

    def _get_deliverable_samples(self, case: Case, workflow: str) -> list[Sample]:
        deliverable_samples: list[Sample] = []
        samples: list[Sample] = self.store.get_samples_by_case_id(case.internal_id)
        for sample in samples:
            bundle_name: str = get_bundle_name(case=case, sample=sample, workflow=workflow)
            bundle_exists: bool = self._bundle_exists(bundle_name=bundle_name, workflow=workflow)
            sample_is_deliverable: bool = self._is_sample_deliverable(sample)
            if bundle_exists and sample_is_deliverable:
                deliverable_samples.append(sample)
        return deliverable_samples

    def _get_version_for_sample(self, sample: Sample, case: Case, workflow: str) -> Version:
        bundle: str = sample.internal_id if workflow == DataDelivery.FASTQ else case.internal_id
        return self.hk_api.last_version(bundle)

    def _get_case_files(self, case: Case, workflow: str) -> list[Path]:
        version: Version = self.hk_api.last_version(case.internal_id)
        sample_ids: set[str] = self._get_sample_ids_for_case(case)

        if not version or not version.files:
            return []

        case_files: list[Path] = []
        for file in version.files:
            if should_include_file_case(file=file, sample_ids=sample_ids, workflow=workflow):
                case_files.append(Path(file.full_path))
        return case_files

    def _get_sample_ids_for_case(self, case: Case) -> set[str]:
        samples: list[Sample] = self.store.get_samples_by_case_id(case.internal_id)
        sample_ids: set[str] = {sample.internal_id for sample in samples}
        return sample_ids

    def _get_sample_files_from_version(
        self,
        version: Version,
        workflow: str,
    ) -> list[Path]:
        """Fetch files for a sample from a version that are tagged with sample tags."""
        sample_files: list[Path] = []
        for file in version.files:
            if should_include_file_sample(file=file, workflow=workflow):
                sample_files.append(Path(file.full_path))
        return sample_files

    def _is_case_deliverable(self, case: Case, workflow: str) -> bool:
        bundle_name: str = case.internal_id
        return self._bundle_exists(bundle_name=bundle_name, workflow=workflow)

    def _is_sample_deliverable(self, sample: Sample) -> bool:
        return (
            sample.sequencing_qc
            or self.deliver_failed_samples
            or sample.application_version.application.is_external
        )

    def _bundle_exists(self, bundle_name: str, workflow: str) -> bool:
        if self.hk_api.last_version(bundle_name):
            return True
        if self.ignore_missing_bundle(workflow):
            return False
        raise SyntaxError(f"Could not find any version for {bundle_name}")

    def ignore_missing_bundle(self, workflow: str) -> bool:
        return workflow in constants.SKIP_MISSING or self.ignore_missing_bundles

    def concatenate_fastqs(self, sample_directory: Path, sample_name: str):
        if self.dry_run:
            return
        forward_out_path = Path(sample_directory, f"{sample_name}_R1.fastq.gz")
        reverse_out_path = Path(sample_directory, f"{sample_name}_R2.fastq.gz")
        self.fastq_file_service.concatenate(
            fastq_directory=sample_directory,
            forward_output=forward_out_path,
            reverse_output=reverse_out_path,
            remove_raw=True,
        )



"""Fixtures for cli balsamic tests."""

import logging
from datetime import datetime
from pathlib import Path
from tempfile import tempdir

import pytest

from cg.apps.gens import GensAPI
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.scout.scoutapi import ScoutAPI
from cg.apps.tb import TrailblazerAPI
from cg.constants.constants import FileFormat
from cg.constants.housekeeper_tags import (
    HK_DELIVERY_REPORT_TAG,
    GensAnalysisTag,
    HkMipAnalysisTag,
)
from cg.io.controller import ReadFile
from cg.meta.deliver import DeliveryAPI
from cg.meta.rsync import RsyncAPI
from cg.meta.upload.scout.uploadscoutapi import UploadScoutAPI
from cg.meta.workflow.mip import MipAnalysisAPI
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.models.scout.scout_load_config import ScoutLoadConfig
from cg.services.fastq_file_service.fastq_file_service import FastqFileService
from cg.store.models import Analysis
from cg.store.store import Store
from tests.meta.upload.scout.conftest import mip_load_config
from tests.mocks.hk_mock import MockHousekeeperAPI
from tests.mocks.madeline import MockMadelineAPI
from tests.mocks.report import MockMipDNAReportAPI
from tests.store_helpers import StoreHelpers

LOG = logging.getLogger(__name__)


@pytest.fixture
def upload_genotypes_hk_bundle(
    case_id: str, timestamp, case_qc_metrics_deliverables: Path, bcf_file: Path
) -> dict:
    """Returns a dictionary in hk format with files used in upload gt process"""
    return {
        "name": case_id,
        "created": datetime.now(),
        "expires": datetime.now(),
        "files": [
            {
                "path": str(case_qc_metrics_deliverables),
                "archive": False,
                "tags": HkMipAnalysisTag.QC_METRICS,
            },
            {"path": str(bcf_file), "archive": False, "tags": ["snv-gbcf", "genotype"]},
        ],
    }


@pytest.fixture
def analysis_obj(
    analysis_store_trio: Store, case_id: str, timestamp: datetime, helpers
) -> Analysis:
    """Return a analysis object with a trio"""
    return analysis_store_trio.get_case_by_internal_id(internal_id=case_id).analyses[0]


@pytest.fixture
def upload_genotypes_hk_api(
    real_housekeeper_api: HousekeeperAPI,
    upload_genotypes_hk_bundle: dict,
    analysis_obj: Analysis,
    helpers,
) -> HousekeeperAPI:
    """Add and include files from upload genotypes hk bundle"""
    helpers.ensure_hk_bundle(real_housekeeper_api, upload_genotypes_hk_bundle)
    hk_version = real_housekeeper_api.last_version(analysis_obj.case.internal_id)
    real_housekeeper_api.include(hk_version)
    return real_housekeeper_api


@pytest.fixture
def upload_gens_hk_bundle(
    case_id: str,
    gens_coverage_path: Path,
    gens_fracsnp_path: Path,
    later_timestamp: datetime,
    sample_id: str,
    timestamp: datetime,
) -> dict:
    """Returns a dictionary in Housekeeper format with files used in upload gens process."""
    return {
        "name": case_id,
        "created": timestamp,
        "expires": later_timestamp,
        "files": [
            {
                "path": gens_coverage_path.as_posix(),
                "archive": False,
                "tags": [sample_id] + GensAnalysisTag.COVERAGE,
            },
            {
                "path": gens_fracsnp_path.as_posix(),
                "archive": False,
                "tags": [sample_id] + GensAnalysisTag.FRACSNP,
            },
        ],
    }


@pytest.fixture
def upload_gens_hk_api(
    case_id: str,
    helpers: StoreHelpers,
    real_housekeeper_api: HousekeeperAPI,
    upload_gens_hk_bundle: dict,
) -> HousekeeperAPI:
    """Add and include files from upload_gens_hk_bundle."""
    helpers.ensure_hk_bundle(store=real_housekeeper_api, bundle_data=upload_gens_hk_bundle)
    hk_version = real_housekeeper_api.last_version(bundle=case_id)
    real_housekeeper_api.include(hk_version)
    return real_housekeeper_api


@pytest.fixture
def upload_gens_context(
    analysis_store_trio: Store,
    base_context: CGConfig,
    gens_api: GensAPI,
    upload_gens_hk_api: HousekeeperAPI,
) -> CGConfig:
    """Create a gens upload context."""
    base_context.gens_api_ = gens_api
    base_context.housekeeper_api_ = upload_gens_hk_api
    base_context.status_db_ = analysis_store_trio
    return base_context


@pytest.fixture
def upload_report_hk_bundle(case_id: str, delivery_report_html: Path, timestamp) -> dict:
    """Returns a dictionary including the delivery report html file"""

    return {
        "name": case_id,
        "created": datetime.now(),
        "files": [
            {"path": str(delivery_report_html), "archive": False, "tags": [HK_DELIVERY_REPORT_TAG]}
        ],
    }


@pytest.fixture
def upload_report_hk_api(
    real_housekeeper_api: HousekeeperAPI,
    upload_report_hk_bundle: dict,
    analysis_obj: Analysis,
    helpers,
) -> HousekeeperAPI:
    """Add and include files from upload reports hk bundle"""

    helpers.ensure_hk_bundle(real_housekeeper_api, upload_report_hk_bundle)
    hk_version = real_housekeeper_api.last_version(analysis_obj.case.internal_id)
    real_housekeeper_api.include(hk_version)
    return real_housekeeper_api


@pytest.fixture
def base_context(
    analysis_store: Store,
    housekeeper_api: HousekeeperAPI,
    upload_scout_api: UploadScoutAPI,
    trailblazer_api: TrailblazerAPI,
    cg_context: CGConfig,
) -> CGConfig:
    """context to use in cli"""
    cg_context.status_db_ = analysis_store
    cg_context.housekeeper_api_ = housekeeper_api
    cg_context.trailblazer_api_ = trailblazer_api
    cg_context.scout_api_ = MockScoutApi()
    cg_context.meta_apis["scout_upload_api"] = upload_scout_api
    cg_context.mip_rd_dna.root = tempdir

    return cg_context


@pytest.fixture
def fastq_context(
    base_context,
    analysis_store: Store,
    housekeeper_api: HousekeeperAPI,
    upload_scout_api: UploadScoutAPI,
    trailblazer_api: TrailblazerAPI,
    cg_context: CGConfig,
) -> CGConfig:
    """Fastq context to use in cli"""

    base_context.delivery_api = DeliveryAPI(
        store=base_context.status_db,
        hk_api=base_context.housekeeper_api,
        pipeline="fastq",
        customers_folder=Path(base_context.delivery_path),
        fastq_file_service=FastqFileService(),
    )
    base_context.meta_apis["rsync_api"] = RsyncAPI(cg_context)
    base_context.trailblazer_api_ = trailblazer_api
    return base_context


@pytest.fixture(scope="function")
def upload_scout_api(housekeeper_api: MockHousekeeperAPI, mip_load_config: ScoutLoadConfig):
    """Return a upload scout api"""
    api = MockScoutUploadApi()
    api.housekeeper = housekeeper_api
    api.config = mip_load_config

    return api


class MockScoutApi(ScoutAPI):
    def __init__(self):
        """docstring for __init__"""

    def upload(self, scout_load_config: Path, force: bool = False):
        """docstring for upload"""
        LOG.info("Case loaded successfully to Scout")


class MockAnalysisApi(MipAnalysisAPI):
    def __init__(self):
        """docstring for __init__"""

    def get_latest_metadata(self, internal_id):
        """docstring for upload"""
        return {}


class MockScoutUploadApi(UploadScoutAPI):
    def __init__(self, **kwargs):
        """docstring for __init__"""
        self.mock_generate_config = True
        self.housekeeper = None
        self.madeline_api = MockMadelineAPI()
        self.analysis = MockAnalysisApi()
        self.config = ScoutLoadConfig(
            delivery_report=Path("path", "to", "delivery-report.html").as_posix()
        )
        self.file_exists = False
        self.lims = MockLims()
        self.missing_mandatory_field = False

    @pytest.fixture(autouse=True)
    def _request_analysis(self, analysis_store_single_case):
        self.analysis = analysis_store_single_case

    def generate_config(self, analysis, **kwargs):
        """Mock the generate config"""
        if self.missing_mandatory_field:
            self.config.vcf_snv = None

        return self.config

    def save_config_file(self, upload_config, file_path):
        """docstring for save_config_file"""
        return

    def add_scout_config_to_hk(self, config_file_path: str, case_id: str, delete: bool = False):
        """docstring for add_scout_config_to_hk"""
        LOG.info("Use mock to upload file")
        if self.file_exists:
            raise FileExistsError("Scout config already exists")


class MockLims:
    """Mock lims fixture"""

    lims = None

    def __init__(self):
        self.lims = self

    @staticmethod
    def lims_samples() -> dict:
        """Return LIMS-like case samples"""
        lims_case: dict = ReadFile.get_content_from_file(
            file_format=FileFormat.JSON,
            file_path=Path("tests", "fixtures", "report", "lims_family.json"),
        )
        return lims_case["samples"]

    def sample(self, sample_id) -> str | None:
        """Returns a lims sample matching the provided sample_id"""
        for sample in self.lims_samples():
            if sample["id"] == sample_id:
                return sample
        return None


@pytest.fixture
def upload_context(cg_context: CGConfig) -> CGConfig:
    analysis_api = MipDNAAnalysisAPI(config=cg_context)
    cg_context.meta_apis["analysis_api"] = analysis_api
    cg_context.meta_apis["report_api"] = MockMipDNAReportAPI(cg_context, analysis_api)
    cg_context.meta_apis["scout_upload_api"] = UploadScoutAPI(
        hk_api=cg_context.housekeeper_api,
        scout_api=cg_context.scout_api,
        madeline_api=cg_context.madeline_api,
        analysis_api=analysis_api,
        lims_api=cg_context.lims_api,
        status_db=cg_context.status_db,
    )

    return cg_context



"""Module to handle time functions"""

import time
from pathlib import Path

from cg.constants.time import SECONDS_IN_A_DAY
from cg.utils.files import get_directory_creation_time_stamp


def get_start_time() -> float:
    """Returns the current time"""
    return time.time()


def get_elapsed_time(start_time: float) -> float:
    """Determines the time elapsed since the start time"""
    return time.time() - start_time


def is_directory_older_than_days_old(directory_path: Path, days_old: int) -> bool:
    """
    Check if the directory is older than the specified number of days.
    """
    days_in_seconds = SECONDS_IN_A_DAY * days_old
    dir_creation_time_stamp: float = get_directory_creation_time_stamp(
        directory_path=directory_path
    )
    return bool(dir_creation_time_stamp < time.time() - days_in_seconds)



"""Test for the time utils functions."""

import time
from pathlib import Path

import mock

from cg.constants.time import TWENTY_ONE_DAYS, TWENTY_ONE_DAYS_IN_SECONDS
from cg.utils.time import is_directory_older_than_days_old


def test_is_directory_older_than_days_old(tmp_path_factory):
    """Test whether a directory is older than a certain number of days."""
    # GIVEN a directory that is created now
    created_now_path: Path = tmp_path_factory.mktemp("created_now")

    # THEN checking whether a directory is older than 21 days returns true
    with mock.patch(
        "time.time",
        return_value=time.time() + TWENTY_ONE_DAYS_IN_SECONDS,
    ):
        assert is_directory_older_than_days_old(
            directory_path=created_now_path, days_old=TWENTY_ONE_DAYS
        )


def test_is_directory_not_older_than_days_old(tmp_path_factory):
    # GIVEN a directory that is created now
    created_now_path: Path = tmp_path_factory.mktemp("created_now")

    # THEN checking whether a directory is older than 21 days returns false
    assert not is_directory_older_than_days_old(
        directory_path=created_now_path, days_old=TWENTY_ONE_DAYS
    )



import logging
from abc import abstractmethod

from pydantic import BaseModel, ConfigDict, Field

from cg.apps.demultiplex.sample_sheet.index import (
    MINIMUM_HAMMING_DISTANCE,
    get_hamming_distance_index_1,
    get_hamming_distance_index_2,
    get_reverse_complement_dna_seq,
    is_dual_index,
    is_padding_needed,
    pad_index_one,
    pad_index_two,
)
from cg.apps.demultiplex.sample_sheet.validators import SampleId
from cg.constants.constants import GenomeVersion
from cg.constants.demultiplexing import (
    CUSTOM_INDEX_TAIL,
    SampleSheetBcl2FastqSections,
    SampleSheetBCLConvertSections,
)
from cg.constants.symbols import EMPTY_STRING
from cg.exc import SampleSheetError
from cg.models.demultiplex.run_parameters import RunParameters

LOG = logging.getLogger(__name__)


class FlowCellSample(BaseModel):
    """Base class for flow cell samples."""

    lane: int
    sample_id: SampleId
    index: str
    index2: str = EMPTY_STRING
    model_config = ConfigDict(populate_by_name=True)

    def separate_indexes(self, is_run_single_index: bool) -> None:
        """Update values for index and index2 splitting the original LIMS dual index."""
        if is_dual_index(self.index):
            index1, index2 = self.index.split("-")
            self.index = index1.strip().replace(CUSTOM_INDEX_TAIL, EMPTY_STRING)
            self.index2 = index2.strip() if not is_run_single_index else EMPTY_STRING

    @abstractmethod
    def process_indexes(self, run_parameters: RunParameters):
        """Update index attributes with the final values for the sample sheet."""
        pass

    @abstractmethod
    def update_barcode_mismatches(
        self, samples_to_compare: list, is_run_single_index: bool, is_reverse_complement: bool
    ) -> None:
        """Update the barcode_mismatches_1 and barcode_mismatches_2 attributes."""
        pass


class FlowCellSampleBcl2Fastq(FlowCellSample):
    """Class that represents a Bcl2Fastq sample."""

    flowcell_id: str = Field("", alias=SampleSheetBcl2FastqSections.Data.FLOW_CELL_ID)
    lane: int = Field(..., alias=SampleSheetBcl2FastqSections.Data.LANE)
    sample_ref: str = Field(
        GenomeVersion.hg19, alias=SampleSheetBcl2FastqSections.Data.SAMPLE_REFERENCE
    )
    index: str = Field(..., alias=SampleSheetBcl2FastqSections.Data.INDEX_1)
    index2: str = Field("", alias=SampleSheetBcl2FastqSections.Data.INDEX_2)
    sample_name: str = Field(..., alias=SampleSheetBcl2FastqSections.Data.SAMPLE_NAME)
    control: str = Field("N", alias=SampleSheetBcl2FastqSections.Data.CONTROL)
    recipe: str = Field("R1", alias=SampleSheetBcl2FastqSections.Data.RECIPE)
    operator: str = Field("script", alias=SampleSheetBcl2FastqSections.Data.OPERATOR)

    sample_id: SampleId = Field(
        ..., alias=SampleSheetBcl2FastqSections.Data.SAMPLE_INTERNAL_ID_BCL2FASTQ
    )
    project: str = Field(..., alias=SampleSheetBcl2FastqSections.Data.SAMPLE_PROJECT_BCL2FASTQ)

    def _pad_indexes_if_necessary(self, run_parameters: RunParameters) -> None:
        index_length: int = len(self.index)
        if is_padding_needed(
            index1_cycles=run_parameters.get_index_1_cycles(),
            index2_cycles=run_parameters.get_index_2_cycles(),
            sample_index_length=index_length,
        ):
            LOG.debug("Padding indexes")
            self.index = pad_index_one(index_string=self.index)
            self.index2 = pad_index_two(
                index_string=self.index2,
                reverse_complement=run_parameters.index_settings.should_i5_be_reverse_complemented,
            )
            return
        LOG.debug(f"Padding not necessary for sample {self.sample_id}")

    def process_indexes(self, run_parameters: RunParameters):
        """Parses, pads and reverse complement the indexes if necessary."""
        reverse_index2: bool = run_parameters.index_settings.should_i5_be_reverse_complemented
        self.separate_indexes(is_run_single_index=run_parameters.is_single_index)
        self._pad_indexes_if_necessary(run_parameters=run_parameters)
        if reverse_index2:
            self.index2 = get_reverse_complement_dna_seq(self.index2)

    def update_barcode_mismatches(
        self, samples_to_compare: list, is_run_single_index: bool, is_reverse_complement: bool
    ) -> None:
        """No updating of barcode mismatch values for Bcl2Fastq samples."""
        LOG.debug(f"No updating of barcode mismatch values for Bcl2Fastq sample {self.sample_id}")


class FlowCellSampleBCLConvert(FlowCellSample):
    """Class that represents a BCLConvert sample."""

    lane: int = Field(..., alias=SampleSheetBCLConvertSections.Data.LANE)
    sample_id: SampleId = Field(..., alias=SampleSheetBCLConvertSections.Data.SAMPLE_INTERNAL_ID)
    index: str = Field(..., alias=SampleSheetBCLConvertSections.Data.INDEX_1)
    index2: str = Field(EMPTY_STRING, alias=SampleSheetBCLConvertSections.Data.INDEX_2)
    override_cycles: str = Field(
        EMPTY_STRING, alias=SampleSheetBCLConvertSections.Data.OVERRIDE_CYCLES
    )
    barcode_mismatches_1: int = Field(
        1, alias=SampleSheetBCLConvertSections.Data.BARCODE_MISMATCHES_1
    )
    barcode_mismatches_2: int | str = Field(
        1, alias=SampleSheetBCLConvertSections.Data.BARCODE_MISMATCHES_2
    )

    def _get_index1_override_cycles(self, len_index1_cycles: int) -> str:
        """Create the index 1 sub-string for the override cycles attribute."""
        len_index_1_sample: int = len(self.index)
        cycles_format: str = f"I{len_index1_cycles};"
        if len_index_1_sample < len_index1_cycles:
            cycles_format = f"I{len_index_1_sample}N{len_index1_cycles - len_index_1_sample};"
        return cycles_format

    def _get_index2_override_cycles(self, len_index2_cycles: int, reverse_cycle: bool) -> str:
        """Create the index 2 sub-string for the override cycles attribute."""
        len_index_2_sample: int = len(self.index2)
        cycles_format: str = f"I{len_index2_cycles};"
        if len_index2_cycles == 0:  # The run was single-index
            cycles_format = EMPTY_STRING
        elif len_index_2_sample == 0:  # The sample was single-index
            cycles_format = f"N{len_index2_cycles};"
        elif len_index_2_sample < len_index2_cycles:
            cycles_format = (
                f"N{len_index2_cycles-len_index_2_sample}I{len_index_2_sample};"
                if reverse_cycle
                else f"I{len_index_2_sample}N{len_index2_cycles - len_index_2_sample};"
            )
        return cycles_format

    def update_override_cycles(self, run_parameters: RunParameters) -> None:
        """Updates the override cycles attribute."""
        reverse_index2_cycles: bool = (
            run_parameters.index_settings.are_i5_override_cycles_reverse_complemented
        )
        read1_cycles: str = f"Y{run_parameters.get_read_1_cycles()};"
        read2_cycles: str = f"Y{run_parameters.get_read_2_cycles()}"
        index1_cycles: str = self._get_index1_override_cycles(run_parameters.get_index_1_cycles())
        index2_cycles: str = self._get_index2_override_cycles(
            len_index2_cycles=run_parameters.get_index_2_cycles(),
            reverse_cycle=reverse_index2_cycles,
        )
        self.override_cycles = read1_cycles + index1_cycles + index2_cycles + read2_cycles

    def _update_barcode_mismatches_1(
        self, samples_to_compare: list["FlowCellSampleBCLConvert"]
    ) -> None:
        """Assign zero to barcode_mismatches_1 if the hamming distance between self.index
        and the index1 of any sample in the lane is below the minimum threshold."""
        for sample in samples_to_compare:
            if self.sample_id == sample.sample_id:
                continue
            if (
                get_hamming_distance_index_1(sequence_1=self.index, sequence_2=sample.index)
                < MINIMUM_HAMMING_DISTANCE
            ):
                LOG.debug(f"Turning barcode mismatch for index 1 to 0 for sample {self.sample_id}")
                self.barcode_mismatches_1 = 0
                break

    def _update_barcode_mismatches_2(
        self,
        samples_to_compare: list["FlowCellSampleBCLConvert"],
        is_reverse_complement: bool,
    ) -> None:
        """Assign zero to barcode_mismatches_2 if the hamming distance between self.index2
        and the index2 of any sample in the lane is below the minimum threshold.
        If the sample is single-indexed, assign 'na'."""
        if self.index2 == EMPTY_STRING and "-" not in self.index:
            LOG.debug(f"Turning barcode mismatch for index 2 to 'na' for sample {self.sample_id}")
            self.barcode_mismatches_2 = "na"
            return
        for sample in samples_to_compare:
            if self.sample_id == sample.sample_id:
                continue
            if (
                get_hamming_distance_index_2(
                    sequence_1=self.index2,
                    sequence_2=sample.index2,
                    is_reverse_complement=is_reverse_complement,
                )
                < MINIMUM_HAMMING_DISTANCE
            ):
                LOG.debug(f"Turning barcode mismatch for index 2 to 0 for sample {self.sample_id}")
                self.barcode_mismatches_2 = 0
                break

    def process_indexes(self, run_parameters: RunParameters):
        """Parse and reverse complement the indexes and updates override cycles."""
        self.separate_indexes(is_run_single_index=run_parameters.is_single_index)
        if run_parameters.index_settings.should_i5_be_reverse_complemented:
            self.index2 = get_reverse_complement_dna_seq(self.index2)
        self.update_override_cycles(run_parameters=run_parameters)

    def update_barcode_mismatches(
        self,
        samples_to_compare: list["FlowCellSampleBCLConvert"],
        is_run_single_index: bool,
        is_reverse_complement: bool,
    ) -> None:
        """Update barcode mismatch attributes comparing to the rest of the samples in the lane."""
        if not samples_to_compare:
            raise SampleSheetError("No samples to compare with to update barcode mismatch values")
        self._update_barcode_mismatches_1(samples_to_compare=samples_to_compare)
        if is_run_single_index:
            LOG.debug("Run is single-indexed, skipping barcode mismatch update for index 2")
            return
        self._update_barcode_mismatches_2(
            samples_to_compare=samples_to_compare, is_reverse_complement=is_reverse_complement
        )



from cg.exc import CaseNotFoundError
from cg.server.dto.delivery_message_response import DeliveryMessageResponse
from cg.services.delivery_message.utils import get_message
from cg.store.store import Store


class DeliveryMessageService:
    def __init__(self, store: Store) -> None:
        self.store = store

    def get_delivery_message(self, case_id: str) -> DeliveryMessageResponse:
        if case := self.store.get_case_by_internal_id(case_id):
            message: str = get_message(case)
            return DeliveryMessageResponse(message=message)
        else:
            raise CaseNotFoundError



from datetime import datetime
from pathlib import Path
from typing import Any

from pydantic import BaseModel, ConfigDict, model_validator

from cg.meta.upload.gisaid.constants import AUTHORS


class FastaFile(BaseModel):
    header: str
    sequence: str


class GisaidAccession(BaseModel):
    log_message: str
    accession_nr: str | None = None
    sample_id: str | None = None

    @model_validator(mode="before")
    @classmethod
    def set_generated_fields(cls, data: Any) -> Any:
        """Constructs the fields that are generated from other fields."""
        if isinstance(data, dict):
            data.setdefault("accession_nr", _parse_accession_nr(data["log_message"]))
            data.setdefault("sample_id", _parse_sample_id_from_log(data["log_message"]))
        return data


def _parse_accession_nr(log_message: str) -> str:
    return log_message.split(";")[-1]


def _parse_sample_id_from_log(log_message: str) -> str:
    return log_message.split("/")[2].split("_")[2]


class UploadFiles(BaseModel):
    csv_file: Path
    fasta_file: Path
    log_file: Path


class GisaidSample(BaseModel):
    model_config = ConfigDict(coerce_numbers_to_str=True)
    case_id: str
    cg_lims_id: str
    submitter: str
    region: str
    region_code: str
    fn: str
    covv_collection_date: str
    covv_subm_sample_id: str
    covv_virus_name: str | None = None
    covv_orig_lab: str | None = None
    covv_type: str | None = "betacoronavirus"
    covv_passage: str | None = "Original"
    covv_location: str | None = None
    covv_host: str | None = "Human"
    covv_gender: str | None = "unknown"
    covv_patient_age: str | None = "unknown"
    covv_patient_status: str | None = "unknown"
    covv_seq_technology: str | None = "Illumina NovaSeq"
    covv_orig_lab_addr: str | None = None
    covv_subm_lab: str | None = "Karolinska University Hospital"
    covv_subm_lab_addr: str | None = "171 76 Stockholm, Sweden"
    covv_authors: str | None = " ,".join(AUTHORS)

    @model_validator(mode="before")
    @classmethod
    def set_generated_fields(cls, data: Any) -> Any:
        """Constructs the fields that are generated from other fields."""
        if isinstance(data, dict):
            data.setdefault("covv_location", _generate_covv_location(data.get("region")))
            data["covv_subm_sample_id"] = _generate_covv_subm_sample_id(
                subm_sample_id=data.get(
                    "covv_subm_sample_id",
                ),
                region_code=data.get("region_code"),
            )
            data.setdefault(
                "covv_virus_name",
                _generate_covv_virus_name(
                    covv_subm_sample_id=data.get("covv_subm_sample_id"),
                    covv_collection_date=data.get("covv_collection_date"),
                ),
            )
        return data


def _generate_covv_location(region: str) -> str:
    return f"Europe/Sweden/{region}"


def _generate_covv_subm_sample_id(subm_sample_id: str, region_code: str) -> str:
    return f"{region_code}_SE100_{subm_sample_id}"


def _generate_covv_virus_name(covv_subm_sample_id: str, covv_collection_date: str) -> str:
    datetime_date: datetime = datetime.strptime(covv_collection_date, "%Y-%m-%d")
    return f"hCoV-19/Sweden/{covv_subm_sample_id}/{datetime_date.year}"



from pathlib import Path

from cg.apps.demultiplex.sample_sheet.sample_models import FlowCellSampleBCLConvert
from cg.apps.demultiplex.sample_sheet.sample_sheet_models import SampleSheet
from cg.apps.demultiplex.sample_sheet.sample_sheet_validator import SampleSheetValidator
from cg.apps.demultiplex.sample_sheet.validators import is_valid_sample_internal_id


def test_get_non_pooled_samples_when_no_samples():
    # GIVEN a sample sheet with no samples
    sample_sheet = SampleSheet(samples=[])

    # WHEN retrieving any non pooled samples
    samples = sample_sheet.get_non_pooled_samples()

    # THEN no samples should be returned
    assert not samples


def test_get_non_pooled_samples_when_multiple_samples_same_lane():
    # GIVEN a sample sheet with multiple samples on the same lane
    sample1 = FlowCellSampleBCLConvert(lane=1, sample_id="ACC123", index="A")
    sample2 = FlowCellSampleBCLConvert(lane=1, sample_id="ACC456", index="A")
    sample_sheet = SampleSheet(samples=[sample1, sample2])

    # WHEN retrieving any non pooled samples
    samples = sample_sheet.get_non_pooled_samples()

    # THEN no samples should be returned
    assert not samples


def test_get_non_pooled_samples_when_single_sample_one_lane():
    # GIVEN a sample sheet with a single sample on one lane
    non_pooled_sample = FlowCellSampleBCLConvert(lane=1, sample_id="ACC123", index="A")
    sample_sheet = SampleSheet(samples=[non_pooled_sample])

    # WHEN retrieving any non pooled samples
    samples = sample_sheet.get_non_pooled_samples()

    # THEN the non pooled sample should be returned
    assert samples == [non_pooled_sample]


def test_get_non_pooled_samples_when_multiple_lanes_one_single():
    # GIVEN a sample sheet with multiple lanes and a single sample in one lane
    sample1 = FlowCellSampleBCLConvert(lane=1, sample_id="ACC123", index="A")
    sample2 = FlowCellSampleBCLConvert(lane=1, sample_id="ACC456", index="A")
    non_pooled_sample = FlowCellSampleBCLConvert(lane=2, sample_id="ACC789", index="A")
    sample_sheet = SampleSheet(samples=[sample1, sample2, non_pooled_sample])

    # WHEN retrieving any non pooled samples
    samples = sample_sheet.get_non_pooled_samples()

    # THEN the non pooled sample should be returned
    assert samples == [non_pooled_sample]


def test_get_sample_internal_ids_from_sample_sheet(
    sample_sheet_validator: SampleSheetValidator,
    novaseq_6000_post_1_5_kits_sample_sheet_object: SampleSheet,
):
    """Test that getting sample internal ids from a sample sheet returns a unique list of strings."""

    # GIVEN a sample sheet with only valid samples

    # WHEN getting the valid sample internal ids
    sample_internal_ids: list[str] = novaseq_6000_post_1_5_kits_sample_sheet_object.get_sample_ids()

    # THEN the returned value is a list
    assert isinstance(sample_internal_ids, list)
    # THEN the list contains strings
    assert isinstance(sample_internal_ids[0], str)
    # THEN the sample internal ids are unique
    assert len(sample_internal_ids) == len(set(sample_internal_ids))
    # THEN the sample internal ids are the expected ones
    for sample_internal_id in sample_internal_ids:
        assert is_valid_sample_internal_id(sample_internal_id=sample_internal_id) is True



import datetime as dt

from cg.apps.tb.api import TrailblazerAPI
from cg.constants.constants import AnalysisType, Workflow
from cg.constants.tb import AnalysisStatus
from cg.exc import CaseNotFoundError
from cg.store.models import Analysis, Case
from cg.store.store import Store


class FastqService:
    def __init__(self, store: Store, trailblazer_api: TrailblazerAPI):
        self.store = store
        self.trailblazer_api = trailblazer_api

    def store_analysis(self, case_id: str) -> None:
        case: Case = self._get_case(case_id)
        self._add_analysis_to_store(case)
        self._add_analysis_to_trailblazer(case)

    def _get_case(self, case_id: str) -> Case:
        if case := self.store.get_case_by_internal_id(case_id):
            return case
        raise CaseNotFoundError(f"Case {case_id} not found.")

    def _add_analysis_to_store(self, case: Case) -> None:
        new_analysis: Analysis = self.store.add_analysis(
            workflow=Workflow.FASTQ,
            completed_at=dt.datetime.now(),
            primary=True,
            started_at=dt.datetime.now(),
            case_id=case.id,
        )
        self.store.session.add(new_analysis)
        self.store.session.commit()

    def _add_analysis_to_trailblazer(self, case: Case) -> None:
        self.trailblazer_api.add_pending_analysis(
            case_id=case.internal_id,
            analysis_type=AnalysisType.OTHER,
            config_path="",
            out_dir="",
            slurm_quality_of_service=case.slurm_priority,
            workflow=Workflow.FASTQ,
            ticket=case.latest_ticket,
        )
        self.trailblazer_api.set_analysis_status(
            case_id=case.internal_id, status=AnalysisStatus.COMPLETED
        )



from datetime import datetime
from pathlib import Path

from cg.apps.sequencing_metrics_parser.models.bcl2fastq_metrics import SampleLaneMetrics
from cg.apps.sequencing_metrics_parser.parsers.bcl2fastq import (
    parse_metrics,
    parse_undetermined_non_pooled_metrics,
)
from cg.apps.sequencing_metrics_parser.sequencing_metrics_calculator import (
    calculate_average_quality_score,
    calculate_q30_bases_percentage,
)
from cg.store.models import SampleLaneSequencingMetrics


def create_bcl2fastq_metrics(bcl2fastq_flow_cell_path: Path) -> list[SampleLaneSequencingMetrics]:
    """Return sequencing metrics for a bcl2fastq flow cell."""
    bcl2fastq_metrics: list[SampleLaneMetrics] = parse_metrics(bcl2fastq_flow_cell_path)
    return convert_to_sequencing_metrics(bcl2fastq_metrics)


def create_bcl2fastq_undetermined_metrics(
    bcl2fastq_flow_cell_path: Path, non_pooled_lane_sample_pairs: list[tuple[str, int]]
) -> list[SampleLaneSequencingMetrics]:
    """Return undetermined sequencing metrics for a bcl2fastq flow cell."""
    undetermined_metrics: list[SampleLaneMetrics] = parse_undetermined_non_pooled_metrics(
        flow_cell_dir=bcl2fastq_flow_cell_path,
        non_pooled_lane_sample_pairs=non_pooled_lane_sample_pairs,
    )
    return convert_to_sequencing_metrics(undetermined_metrics)


def convert_to_sequencing_metrics(
    bcl2fastq_metrics: list[SampleLaneMetrics],
) -> list[SampleLaneSequencingMetrics]:
    """Convert the raw bcl2fastq metrics to SampleLaneSequencingMetrics."""
    return [create_sequencing_metric(metric) for metric in bcl2fastq_metrics]


def create_sequencing_metric(
    bcl2fastq_metric: SampleLaneMetrics,
) -> SampleLaneSequencingMetrics:
    """Generates a SampleLaneSequencingMetrics based on the provided raw results."""

    sample_base_percentage_passing_q30: float = calculate_q30_bases_percentage(
        q30_yield=bcl2fastq_metric.total_yield_q30,
        total_yield=bcl2fastq_metric.total_yield,
    )
    sample_base_mean_quality_score: float = calculate_average_quality_score(
        total_quality_score=bcl2fastq_metric.total_quality_score,
        total_yield=bcl2fastq_metric.total_yield,
    )

    return SampleLaneSequencingMetrics(
        flow_cell_name=bcl2fastq_metric.flow_cell_name,
        flow_cell_lane_number=bcl2fastq_metric.lane,
        sample_internal_id=bcl2fastq_metric.sample_id,
        sample_total_reads_in_lane=bcl2fastq_metric.total_reads,
        sample_base_percentage_passing_q30=sample_base_percentage_passing_q30,
        sample_base_mean_quality_score=sample_base_mean_quality_score,
        created_at=datetime.now(),
    )
